"""Publishable models.

Models that are dynamically displayed and hidden.
"""
__all__ = [
    'PublishExpireMixin',
]

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from .managers import PublishExpireQuerySet


class PublishExpireMixin(models.Model):
    """A time framed model that can be published at a future date and expired.

    Attributes
    ----------
    publish_dt : models.DateTimeField
        Instance will be displayed after the time specified.
        Defaults to the current date and time.
    expiry_dt : models.DateTimeField
        Instance will be hidden after the time specified.
        Value of ``None`` means the instance will never expire.

    Notes
    -----
    ``expiry_dt`` should be more than ``publish_dt``, validation is not handled
    by this mixin.

    """
    objects = PublishExpireQuerySet.as_manager()

    publish_dt = models.DateTimeField(
        verbose_name=_("Publish date"),
        help_text=_("Will only be visible after this time."),
        default=timezone.now,
        blank=True)
    expiry_dt = models.DateTimeField(
        verbose_name=_("Expiry date"),
        help_text=_("Will be hidden after this time. Will not expire if left blank"),
        default=None,
        blank=True,
        null=True)

    class Meta:
        abstract = True



"""Exception utils."""


def raise_truthy(d, k, exception=None):
    """Raises exception if value is truthy.

    Parameters
    ----------
    d : dict
    k : str
    exception : Exception, optional

    Returns
    -------
    None

    Raises
    ------
    ValueError
        When dict value is truthy and a custom exception is not passed.

    """
    val = d.get(k)
    if (val):
        raise exception or ValueError(k)
    return val


def raise_falsey(d, k, exception=None):
    """Return value from dict if truthy, raises otherwise.

    If an exception is provided, it will be the exception raised.

    Parameters
    ----------
    d : dict
    k : str
    exception : Exception, optional

    returns
    -------
    any
        Returns the dict value

    Raises
    ------
    ValueError
        When dict value is falsey and a custom exception is not passed.

    """
    val = d.get(k)
    if not val:
        raise exception or ValueError(k)
    return val


def raise_none(d, k, exception=None):
    """Return value from dict if not None, raises otherwise.

    If an exception is provided, it will be the exception raised.

    Parameters
    ----------
    d : dict
    k : str
    exception : Exception, optional

    returns
    -------
    any
        Returns the dict value

    Raises
    ------
    ValueError
        When dict value is falsey and a custom exception is not passed.

    """
    val = d.get(k)
    if val is None:
        raise exception or ValueError(k)
    return val



"""Auto version serializers."""
from rest_framework import serializers


class AutoIntVersionSerializer(serializers.Serializer):
    version = serializers.IntegerField(read_only=True)

    class Meta:
        fields = ('version',)



"""Util exceptions.

To be used with the utils.views classes.
Raising these exceptions within the view will return the appropriate
response.
"""
__all__ = [
    'APIException',
    'APIBadRequestException',
    'APIUnauthorizedException',
    'APIForbiddenException',
    'APINotFoundException',
    'APITooManyRequestsException',
]

from rest_framework.response import Response


class APIException(Exception):
    """Bad request exception.

    Raising this exception will return a bad request response,
    with the relevant status code and description.

    Parameters
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int
        HTTP status code

    Attributes
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int
        HTTP status code

    """
    def __init__(self, status_code, status_description, http_code):
        self.status_code = status_code
        self.status_description = status_description
        self.http_code = http_code

    def as_tuple(self):
        """Exception info as a tuple for easier unpacking handling.

        Returns
        -------
        tuple
            Tuple of status code, status description and HTTP status code.

        """
        return (self.status_code, self.status_description, self.http_code)

    def as_context(self):
        """Exception info as required in response context.

        Returns
        -------
        dict
            Exception info as context dictionary.

        """
        return {
            'status_cd': self.status_code,
            'status_desc': self.status_description,
        }

    def as_response(self):
        """Exception info as response.

        Can be returned directly from a view.

        Returns
        -------
        Response
            DRF Response.

        """
        return Response(self.as_context(), status=self.http_code)


class APIBadRequestException(APIException):
    """API exception for object not found instances.

    Making use of this exception class means you can catch only this
    exception instead of needing to check the exceptions http_code.

    Parameters
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int, optional
        HTTP status code (defaults to 400)

    Attributes
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int
        HTTP status code

    """
    def __init__(self, status_code, status_description):
        super().__init__(status_code, status_description, http_code=400)


class APIUnauthorizedException(APIException):
    """API exception for object not found instances.

    Making use of this exception class means you can catch only this
    exception instead of needing to check the exceptions http_code.

    Parameters
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int, optional
        HTTP status code (defaults to 401)

    Attributes
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int
        HTTP status code

    """
    def __init__(self, status_code, status_description):
        super().__init__(status_code, status_description, http_code=401)


class APIForbiddenException(APIException):
    """API exception for object not found instances.

    Making use of this exception class means you can catch only this
    exception instead of needing to check the exceptions http_code.

    Parameters
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int, optional
        HTTP status code (defaults to 403)

    Attributes
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int
        HTTP status code

    """
    def __init__(self, status_code, status_description):
        super().__init__(status_code, status_description, http_code=403)


class APINotFoundException(APIException):
    """API exception for object not found instances.

    Making use of this exception class means you can catch only this
    exception instead of needing to check the exceptions http_code.

    Parameters
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int, optional
        HTTP status code (defaults to 404)

    Attributes
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int
        HTTP status code

    """
    def __init__(self, status_code, status_description):
        super().__init__(status_code, status_description, http_code=404)


class APITooManyRequestsException(APIException):
    """API exception for object not found instances.

    Making use of this exception class means you can catch only this
    exception instead of needing to check the exceptions http_code.

    Parameters
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int, optional
        HTTP status code (defaults to 429)

    Attributes
    ----------
    status_code : str
        JSON error status code
    status_description : str
        JSON error description
    http_code : int
        HTTP status code

    """
    def __init__(self, status_code, status_description):
        super().__init__(status_code, status_description, http_code=429)



"""Soft delete managers."""
__all__ = [
    "SoftDeleteQuerySet",
    "SoftDeleteManager",
]

from django.db.models import Manager, QuerySet
from django.utils import timezone


class SoftDeleteQuerySet(QuerySet):
    """Soft delete queryset.

    Overrides the default delete behavior to soft delete instances.
    """

    def delete(self, *, hard=False):
        """Soft delete all instances in the queryset.

        Parameters
        ----------
        hard : bool
            Hard delete the items if True.
            Must be passed as a kwarg.

        Returns
        _______
        None
            Return value should not be relied on.

        """
        if hard:
            return super().delete()
        return super().update(deleted_on=timezone.now())


class SoftDeleteManager(Manager):
    """Soft delete model manager.

    Attributes
    ----------
    show_deleted : bool
        Whether to return soft deleted items.

    """

    def __init__(self, show_deleted=False, *args, **kwargs):
        self.show_deleted = show_deleted
        super().__init__(*args, **kwargs)

    def get_queryset(self):
        """Get soft delete queryset."""
        qs = SoftDeleteQuerySet(self.model)
        if not self.show_deleted:
            qs = qs.filter(deleted_on=None)
        return qs

    def delete(self, *, hard=False):
        """Override delete to do soft deletions by default.

        Parameters
        ----------
        hard : bool
            Hard delete the items if True.
            Must be passed as a kwarg.

        """
        return self.get_queryset().delete(hard=hard)



"""Util middlewares."""
import logging

from django.conf import settings
from django.http import JsonResponse


# Share same logger
logger = logging.getLogger('django.request')


class ExceptionMiddleware:
    """Jewel styled response for unhandled errors.

    Instead of returning a HTML response, return a Jewel styled JSON response.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_exception(self, request, exception):
        """Return a JSON response for unhandled errors instead of HTML.

        If DEBUG is enabled, show a traceback instead.
        """
        if settings.DEBUG:
            return None

        logger.error(
            "Internal Server Error: {0}".format(request.path),
            exc_info=True,
            extra={'status_code': 500, 'request': request}
        )

        return JsonResponse({
            "status_cd": "9999",
            "status_desc": "Server Error",
        }, status=500)



"""Util views.

Reusable Django Rest Framework based views for resource based endpoints.

"""
__all__ = [
    'APIView',
    'ListModelView',
    'CreateListModelView',
    'ModelView',
]

from django.http import Http404, JsonResponse
from django.core.exceptions import PermissionDenied
from rest_framework.views import APIView as _APIView, set_rollback
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework import status
from rest_framework.exceptions import APIException as _APIException

from .mixins import (ListModelMixin, RetrieveModelMixin, UpdateModelMixin,
                     DestroyModelMixin, CreateModelMixin)
from .exceptions import APIException


def exception_handler(exc, context):
    """Custom CaptureV2 API error handler."""
    response = None
    if isinstance(exc, APIException):
        response = exc.as_response()
    elif isinstance(exc, Http404):
        response = Response({
            "status_cd": "9999",
            "status_desc": "Not found."
        }, status=status.HTTP_404_NOT_FOUND)
    elif isinstance(exc, PermissionDenied):
        response = Response({
            "status_cd": "9999",
            "status_desc": "Permission denied."
        }, status=status.HTTP_403_FORBIDDEN)
    elif isinstance(exc, _APIException):
        response = Response({
            "status_cd": "9999",
            "status_desc": exc.detail,
        }, status=exc.status_code)

    if response:
        set_rollback()
    return response


class APIView(_APIView):
    """Generic APIView. Should be used for non resource based endpoints.

    Custom version of rest_framework's APIView. Introduces conveniences for
    working with Jewel's API response format.

    Attributes
    ----------
    success_code : str
        Success code of a successful response
    success_description : str
        Success description of a successful response

    """
    success_code = "0000"
    success_description = "SUCCESSFUL"
    success_http_code = status.HTTP_200_OK

    def get_exception_handler(self):
        return exception_handler

    def success(self, data=None, status_code=None, status_description=None, http_code=None):
        """Return a successful JSON response.

        Parameters
        ----------
        data : dict, optional
            Additional data to send.
        status_code : str, optional
            JSON 'status_cd' value. defaults to "0000"
        status_description : str, optional
            JSON 'status_desc' value. defaults to "SUCCESSFUL"
        http_code : int, optional
            HTTP status code. defaults to 200

        """
        if data is None:
            data = {}
        if status_code is None:
            status_code = self.success_code
        if status_description is None:
            status_description = self.success_description
        if http_code is None:
            http_code = self.success_http_code

        context = {
            "status_cd": status_code,
            "status_desc": status_description,
            **data,
        }
        return Response(context, status=http_code)

    def failure(self, status_code="9999", status_description="", http_code=400, data=None):
        """Return a failure JSON response.

        Parameters
        ----------
        status_code : str, optional
            JSON 'status_cd' value. defaults to "9999"
        status_description : str, optional
            JSON 'status_desc' value. defaults to ""
        http_code : int, optional
            HTTP status code. defaults to 400
        data : dict, optional
            Additional data to send.

        Note
        ----
        ``data`` is the last pos arg in order to preserve backwards compatibility.

        """
        if data is None:
            data = {}
        return JsonResponse({
            "status_cd": status_code,
            "status_desc": status_description,
            **data,
        }, status=http_code)


class _BaseResourceView(APIView, GenericAPIView):
    """Base API view, not intended to be used directly."""

    def get_serializer(self, *args, reading=False, **kwargs):
        serializer_class = self.get_serializer_class(reading)
        kwargs['context'] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self, reading):
        return self.serializer_class


class ListModelView(ListModelMixin, _BaseResourceView):
    """Model list API endpoint view.

    In order to make use of this class the following attributes must be specified
    in the inheriting class::

        serializer_class : rest_framework.serializers.Serializer
            The serializer used to serialize the instance.
        queryset : django.db.models.Queryset
            Queryset of instances too return.
        response_key : str
            Key to return data.

    """

    def get_context(self):
        """Context by GET handler.

        Provided for easier overriding.
        """
        return self.list(self.request)

    def get(self, request):
        """Handle get requests."""
        return self.success(self.get_context())


class CreateListModelView(CreateModelMixin, ListModelView):
    """Model list and create API endpoint view.

    Additional Required attributes (See ListModelView)::

        create_response_key : str
            Key to return created instance data.

    See ``ListModelView`` for usage
    """

    def post_context(self):
        """Context used by POST handler.

        Provided for easier overriding.
        """
        return self.create(self.request)

    def post(self, request):
        """Handle post request."""
        return self.success(self.post_context())


class ModelView(UpdateModelMixin, RetrieveModelMixin, DestroyModelMixin, _BaseResourceView):
    """Model get and edit API endpoint view.

    In order to make use of this class the following attributes must be specified
    in the inheriting class:

    serializer_class : rest_framework.serializers.Serializer
        The serializer used to serialize the instance.
    queryset : django.db.models.Queryset
        Queryset of instances too return.
    lookup_field : str
        The model field that should be used to for performing object lookup of
        individual model instances.
    lookup_kwarg : str, optional
        The URL keyword argument that should be used for object lookup. The URL
        conf should include a keyword argument corresponding to this value.
        If unset this defaults to using the same value as lookup_field.
    not_found_code : str
        Status Code returned when object is not found
    not_found_description : str
        Status description returned when object is not found
    error_code_filed_mapping : dict
        See utils.mixins.UpdateModelMixin

    """

    def get_context(self):
        """Context used by GET handler."""
        return self.retrieve(self.request)

    def get(self, request, *args, **kwargs):
        """Handle GET request."""
        return self.success(self.get_context())

    def post_context(self):
        """Context used by POST handler."""
        return self.update(self.request)

    def post(self, request, *args, **kwargs):
        """Handle POST request."""
        return self.success(self.post_context())

    def validate_delete(self, request, *args, **kwargs):
        """Override to provide custom exception handling.

        Raise an APIException to return bad requests.

        NOTES
        -----
        Depracated!
        Look into handling with Serializer level validation instead.

        """
        pass

    def delete(self, request, *args, **kwargs):
        """Delete request handling."""
        self.validate_delete(request, *args, **kwargs)
        self.destroy(request)

        return self.success()



"""Testing utils."""
import json

from django.test.client import Client


class JSONClient(Client):
    """Django client extended to handle json.

    Sets headers for JSON handling and serializes python data into JSON.

    Attributes
    ----------
    JSON_MIME : str
        JSON Mime type.
    """
    JSON_MIME = "application/json"

    def _json_call(self, path, data=None, method="post", **kwargs):
        """Automatically serializes python to JSON and the appropriate headers.

        Parameters
        ---------
        path : str
        data : dict, list
        method : str
            Https method to use.

        Returns
        -------
        Response
            The same response type the original client returns.

        """
        fn = getattr(super(), method.lower())
        payload = json.dumps(data) if data else None
        return fn(path, payload, content_type=self.JSON_MIME, HTTP_ACCEPT=self.JSON_MIME, **kwargs)

    def get(self, *args, **kwargs):
        """Request a response from the server using GET.

        Refer to ``django.test.client.Client.get`` for accepted params.
        """
        return self._json_call(*args, method="get", **kwargs)

    def post(self, *args, **kwargs):
        """Request a response from the server using POST.

        Refer to ``django.test.client.Client.post`` for accepted params.
        """
        return self._json_call(*args, method="post", **kwargs)

    def put(self, *args, **kwargs):
        """Request a response from the server using PUT.

        Refer to ``django.test.client.Client.put`` for accepted params.
        """
        return self._json_call(*args, method="put", **kwargs)

    def patch(self, *args, **kwargs):
        """Request a response from the server using PATCH.

        Refer to ``django.test.client.Client.patch`` for accepted params.
        """
        return self._json_call(*args, method="patch", **kwargs)

    def delete(self, *args, **kwargs):
        """Request a response from the server using DELETE.

        Refer to ``django.test.client.Client.delete`` for accepted params.
        """
        return self._json_call(*args, method="delete", **kwargs)



"""Created updated utils package.

Examples
--------
TODO

"""
import logging

from .models import *  # noqa
try:
    import rest_framework  # noqa
except ImportError:
    logging.info("DRF not found, skipping DRF imports")
else:
    from .serializers import *  # noqa



"""Auto version admin."""
from django.contrib import admin


class AutoIntVersionAdmin(admin.ModelAdmin):
    """Add version into list_display and as readonly."""

    def get_list_display(self, request):
        list_display = list(super().get_list_display(request)) + [
            'version',
        ]
        return list_display

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super().get_readonly_fields(request, obj)) + [
            'version',
        ]
        return readonly_fields



"""Publishable managers and querysets."""
__all__ = [
    'PublishExpireQuerySet',
    'PublishExpireManager'
]

from django.db import models
from django.utils import timezone


class PublishExpireQuerySet(models.QuerySet):
    """Publish expire queryset."""

    def expired(self):
        """Instances past their expiry date."""
        return self.filter(expiry_dt__lt=timezone.now())

    def published(self):
        """Instances that are not expired or pending."""
        now = timezone.now()
        after_publish_date_q = models.Q(publish_dt__lt=now)
        before_expiry_q = models.Q(expiry_dt__gt=now) | models.Q(expiry_dt=None)
        return self.filter(after_publish_date_q & before_expiry_q)

    def pending(self):
        """Instances that are awaiting to be published."""
        return self.filter(publish_dt__gt=timezone.now())


class PublishExpireManager(models.Manager):
    """Publish expire model manager."""

    def get_queryset(self):
        """Get publish expire queryset."""
        return PublishExpireQuerySet(self.model)

    def expired(self):
        return self.get_queryset().expired()

    def published(self):
        return self.get_queryset().published()

    def pending(self):
        return self.get_queryset().pending()



"""Util fields."""
from django.db import models
from django.utils.translation import ugettext_lazy as _


class AutoIntVersionField(models.PositiveIntegerField):
    """An autoincrement on save integer field.

    Meant to keep track of an instance's "version" based on how many
    times it has been saved.

    Does not check if any of the fields actually change, will increment
    on any ``save()``.

    Attributes
    ----------
    default : int, optional
        Initial version number to use for newly created objects.
    editable : bool, optional
        Defaults to un-editable.

    """
    description = _("Auto increment on save integer field")

    def __init__(self, *args, **kwargs):
        kwargs['default'] = kwargs.get('default', 1)
        kwargs['editable'] = kwargs.get('editable', False)
        super().__init__(*args, **kwargs)

    def pre_save(self, model_instance, add):
        """Handle auto incrementing.

        Parameters
        ----------
        model_instance
            Instance of a model using this field.
            If ``model_instance.update_version = False``, version
            will notincrement.
        add : bool
            True if instance is being created,
            False if instance is being updated (existing)

        Returns
        -------
        int
            New value to set when saving.

        Notes
        -----
        Faced complications when attempting to use F expression to update
        the field.

        """
        value = getattr(model_instance, self.attname)
        if not getattr(model_instance, 'update_version', True):
            return value

        if not add:
            value += 1
        setattr(model_instance, self.attname, value)
        return value



"""Util model mixins."""
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .fields import AutoIntVersionField


class AutoIntVersionMixin(models.Model):
    """Adds an autoincrement on save int field, ``version``.

    Version will update on every save regardless whether field values actually
    change or note.

    If you need to save without incrementing the version field either pass
    ``.save(update_version=False)`` or set ``update_version=False`` directly
    on the instance.

    Attributes
    ----------
    version : AutoIntVersionField
        Check ``AutoIntVersionField`` for default values.

    """
    version = AutoIntVersionField(verbose_name=_("Version"))

    class Meta:
        abstract = True

    def save(self, **kwargs):
        """Detect if update_version is either passed as kwargs or as a prop on model."""
        self.update_version = kwargs.pop('update_version', getattr(self, 'update_version', True))
        super().save(**kwargs)



"""Creted updated serializers."""
from rest_framework import serializers


class CreatedUpdatedSerializer(serializers.Serializer):
    created_dt = serializers.DateTimeField(read_only=True)
    created_by = serializers.CharField(read_only=True)
    updated_dt = serializers.DateTimeField(read_only=True)
    updated_by = serializers.CharField(read_only=True)

    class Meta:
        fields = ('created_dt', 'created_by',
                  'updated_dt', 'updated_by', )



"""Created updated util models."""
__all__ = [
    "CreatedUpdatedMixin",
]

from django.db import models
from django.utils.translation import ugettext_lazy as _


class CreatedUpdatedMixin(models.Model):
    """Mixin to add created, edited timestamp and creator/editor name.

    Created by and updated by is not a foreign key to a user object, just a
    ``django.db.models.CharField`.

    Ensuring that ``created_by`` is not edited after instance creation is the
    developers responsibility.

    """
    created_dt = models.DateTimeField(
        verbose_name=_("Created"),
        auto_now_add=True)
    created_by = models.CharField(
        verbose_name=_("Created by"),
        max_length=225,)
    updated_dt = models.DateTimeField(
        verbose_name=_("Updated"),
        auto_now=True)
    updated_by = models.CharField(
        verbose_name=_("Updated by"),
        max_length=225,)

    class Meta:
        abstract = True



"""Util mixins.

These mixins are designed to work alongside rest_framework's GenericAPIView.

Notes
-----
GenericAPIView must ALWAYS be the last class when inheriting.

"""
__all__ = [
    'ListModelMixin',
    'CreateModelMixin',
    'RetrieveModelMixin',
    'UpdateModelMixin',
    'DestroyModelMixin',
]

from django.http import Http404

from .exceptions import APIBadRequestException, APINotFoundException


class _ResponseKeyMixin:
    """Response data will be nested in a key with this name.

    Attributes
    ----------
    response_key : str
        Key to nest response data under. Defaults to `data`
    """
    response_key = 'data'

    def get_response_key(self):
        assert self.response_key is not None, "response_key not set on {0}".format(self.__class__.__name__)
        return self.response_key


class _ResponseNotFoundMixin:
    """Adds an APIException handler to get_object.

    When get_object fails to find an instance, raises an APIException.

    Attributes
    ----------
    not_found_code : str
        The status code value given to the raised exception.
    not_found_description : str
        The status description given to the raised exception.

    """
    not_found_code = "0001"
    not_found_description = "Instance not found"

    def get_object(self):
        """Get the instance, raise APIException if not found."""
        try:
            return super().get_object()
        except Http404:
            raise APINotFoundException(self.not_found_code, self.not_found_description)


class CreateModelMixin:
    """Adds a `.create()` method"""
    create_response_key = 'data'
    error_code_field_mapping = {}

    def get_error_code(self, field_name):
        """Retrieve the error code of a field, based on the mapping set.

        Returns 9999 if the field isn't found in the mapping.

        Attributes
        ----------
        field_name : str
            Field name

        """
        return self.error_code_field_mapping.get(field_name, '9999')

    def perform_create(self, serializer):
        serializer.save()

    def create(self, request):
        """Create a new instance."""
        serializer = self.get_serializer(data=request.data, reading=False)
        if not serializer.is_valid():
            for field_name, errors in serializer.errors.items():
                error_code = self.get_error_code(field_name)
                error_message = "Invalid \"{0}\": {1}".format(field_name, errors[0])
                raise APIBadRequestException(error_code, error_message)

        self.perform_create(serializer)
        serializer = self.get_serializer(instance=serializer.data, reading=True)

        ret = {
            self.create_response_key: serializer.data
        }
        return ret


class ListModelMixin(_ResponseKeyMixin):
    """Similar to ListModel Mixin but nests the result under a key.

    Attributes
    ----------
    response_key : str
        See `_ResponseKeyMixin`
    table_headers : list
        List of dictionaries.
        title will be the label shown on the front-end.
        (field)name is the key the value is stored in.

        Sample format::

            {
                "title": "Key", "name": "key",
                "title": "Value", "name": "value",
                "title": "Last Updated By", "name": "last_updated",
            }

    """
    response_key = None
    table_headers = None

    def get_data(self):
        queryset = self.get_queryset()
        serializer = self.get_serializer(queryset, many=True, reading=True)
        return serializer.data

    def list(self, request):
        """Return a dictionary with serialized objects under a specified key.

        Parameters
        ----------
        request : Django Request
            Used to handle filtering.

        Returns
        -------
        dict
            The serialized data is nested within a key.

        Notes
        -----
        Filtering is not yet handled.

        """
        ret = {
            self.get_response_key(): self.get_data()
        }
        return ret


class RetrieveModelMixin(_ResponseKeyMixin, _ResponseNotFoundMixin):
    """Adds a `.retrieve()` method.

    Returns the instance's serialized data nested under the specified key.

    Attributes
    ----------
    not_found_code : str
        See `_ResponseNotFoundMixin`
    not_found_description : str
        See `_ResponseNotFoundMixin`

    """
    not_found_code = "0001"
    not_found_description = "Instance not found."

    def retrieve(self, request):
        """Return the object serialized as a dictionary."""
        instance = self.get_object()

        serializer = self.get_serializer(instance, reading=True)

        ret = {
            self.get_response_key(): serializer.data
        }
        return ret


class UpdateModelMixin(_ResponseKeyMixin, _ResponseNotFoundMixin):
    """Adds an `.update()` method.

    Attributes
    ----------
    error_code_field_mapping : dict
        A mapping of field names and error codes.
        The keys should be a string of a fieldname.
        Values should be the error code as a string.

    """
    error_code_field_mapping = {}

    def get_error_code(self, field_name):
        """Retrieve the error code of a field, based on the mapping set.

        Returns 9999 if the field isn't found in the mapping.

        Attributes
        ----------
        field_name : str
            Field name

        """
        return self.error_code_field_mapping.get(field_name, '9999')

    def update(self, request):
        """Update the instance from request data and return the new serialized instance.

        If validation fails, responds with the first error in the list.
        """
        instance = self.get_object()

        serializer = self.get_serializer(instance, data=request.data, reading=False)
        if not serializer.is_valid():
            for field_name, errors in serializer.errors.items():
                error_code = self.get_error_code(field_name)
                error_message = "Invalid \"{0}\": {1}".format(field_name, errors[0])
                raise APIBadRequestException(error_code, error_message)
        obj = serializer.save()

        serializer = self.get_serializer(instance=obj, reading=True)
        ret = {
            self.get_response_key(): serializer.data
        }
        return ret


class DestroyModelMixin(_ResponseNotFoundMixin):
    """Adds a `.destroy()` method for deleting instances."""

    def destroy(self, request):
        """Delete the found instance."""
        instance = self.get_object()
        instance.delete()



"""Soft delete model mixins."""
__all__ = [
    "SoftDeleteMixin",
]

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from .managers import SoftDeleteManager


class SoftDeleteMixin(models.Model):
    """Adds soft deletion to a model.

    Override default delete behavior to soft delete.
    Hard deletes can be done by passing in ``hard=True`` as a kwarg.
    Default manager will not return soft deleted items.

    Soft deletion does not handle any cascading behavior with children.
    This behavior will have to be handled by further overriding the delete
    method.

    Attributes
    ----------
    objects : SoftDeleteManager
        Overrides the default manager to exclude 'deleted' items.
    all_objects : SoftDeleteManager
        Manager that returns all items.
    deleted_on : django.db.models.DateTimeField
        Field used to track soft deleted models.
        Null/None value means model has not been deleted.
        A datetime value indicates when the item was soft deleted.

    Notes
    -----
    May want to look into auto handling cascading soft delete events.
    Especially if children also extend the ``SoftDeleteMixin``.

    """
    objects = SoftDeleteManager()
    all_objects = SoftDeleteManager(show_deleted=True)

    deleted_on = models.DateTimeField(
        verbose_name=_("Deleted on"),
        null=True,
        default=None,
        editable=False)

    class Meta:
        abstract = True

    def delete(self, using=None, *args, hard=False, **kwargs):
        """Override to do soft deletions by default.

        Parameters
        ----------
        hard : bool
            Hard deletes the instance.
            Can only be passed as a kwarg.

        Returns
        -------
        None
            On hard deletes, the usual dict for deletions are returned.
            But soft deletions do not return anything.
            Should not rely on the return value when deleting.

        """
        if hard:
            return super().delete(using=using, *args, **kwargs)

        self.deleted_on = timezone.now()
        self.save(using=using)
